"""
This is a custom and lightweight validation module for Panda Dataframes. There are no
dependencies. It is inspired by the Pandera API and replicates the way JSON schemas are
used to define how a pd.DataFrame is validated. e.g.:

schema = DataFrameSchema(
    "<name of validator>",
    {
        "requiredcolumn1": Column(stringlike),
        "requiredcolumn2": Column(numeric),
        "requiredcolumn3": Column(int64, checks=Check("!=", 0)),
        "requiredcolumn4": Column(float),
        "requiredcolumn5": Column(float, checks=Check("<", "requiredcolumn4")),
    },
)

calling schema.validate(dataframe_to_be_validated) will print warnings for missing
columns, wrong datatypes and failed custom checks.
"""

from geost.utils import COMPARISON_OPERATORS, warn_user

raise_error = False


class ValidationError(Exception):
    pass


class Numeric:
    associated_types: tuple = (
        int,
        float,
        "int64",
        "float64",
        "Int64",
        "Int32",
        "Float64",
        "Float32",
    )

    def __repr__(self):
        return "numeric type"

    def __eq__(self, other):
        return other in self.associated_types


numeric = Numeric()


class StringLike:
    associated_types: tuple = (str, "object", "string")

    def __repr__(self):
        return "stringlike type"

    def __eq__(self, other):
        return other in self.associated_types


stringlike = StringLike()


class DataFrameSchema:
    def __init__(self, name: str, json_schema):
        self.name = name
        self.schema = json_schema
        self.validationerrors = []
        self.column_name = ""
        self.column_validation_parameters = None
        self.dataframe = None

    def _validate_dtype(self):
        column_dtype = self.dataframe[self.column_name].dtype
        if not self.column_validation_parameters.required_dtype == column_dtype:
            self.validationerrors.append(
                f'During {self.name}: datatype in column "{self.column_name}" is '
                f'"{column_dtype}", but is required to be '
                f'"{self.column_validation_parameters.required_dtype}"'
            )
            return False
        return True

    def _validate_checks(self):
        for check in self.column_validation_parameters.checks:
            success, reports = check.check(self.dataframe, self.column_name)
            if not success:
                self.validationerrors.append(
                    f'During {self.name}: data in column "{self.column_name}" '
                    f'failed check "{check}" for {len(reports)} rows: {reports}'
                )
                return False
            return True

    def validate(self, dataframe):
        self.dataframe = dataframe
        for self.column_name, self.column_validation_parameters in self.schema.items():
            if self.column_name not in dataframe.columns:
                self.validationerrors.append(
                    f'During {self.name}: required column "{self.column_name}" is '
                    "missing"
                )
            else:
                passed_dtype_check = self._validate_dtype()
                if passed_dtype_check and self.column_validation_parameters.checks:
                    self._validate_checks()

        if len(self.validationerrors) >= 1:
            if raise_error:
                raise ValidationError(("\n").join(self.validationerrors))
            else:
                self.warn_user()

    @warn_user
    def warn_user(self):
        print(("\n").join(self.validationerrors))


class Check:
    def __init__(self, operator: str, reference_value, report_by="index"):
        self.operator_str = operator
        self.operator = COMPARISON_OPERATORS[operator]
        self.reference = reference_value
        self.report_by = report_by

    def __repr__(self):
        return f"{self.operator_str} {self.reference}"

    def check(self, dataframe, column):
        if self.reference in dataframe.columns:
            reference = dataframe[self.reference]
        else:
            reference = self.reference

        comparison = self.operator(dataframe[column], reference)
        if not all(comparison):
            invalid_indices = dataframe[~comparison].index
            if self.report_by in dataframe.columns:
                return False, dataframe.loc[invalid_indices, self.report_by].values
            else:
                return False, invalid_indices
        else:
            return True, []


class Column:
    def __init__(self, required_dtype, checks=None):
        self.required_dtype = required_dtype
        if checks is not None:
            if checks.__class__ == Check:
                self.checks = [checks]
            elif len(checks) > 1:
                self.checks = checks
        else:
            self.checks = None

    def __repr__(self):
        return f"Column of type {self.required_dtype}"



# class CptCollection(PointDataCollection):
#     """
#     Class for collections of CPT data.

#     Users must use the reader functions in
#     :py:mod:`~geost.read` to create collections. The following readers generate CPT
#     objects:

#     :func:`~geost.read.read_sst_cpts`, :func:`~geost.read.read_gef_cpts`

#     Args:
#         data (pd.DataFrame): Dataframe containing borehole/CPT data.

#         vertical_reference (str): Vertical reference, see
#          :py:attr:`~geost.base.PointDataCollection.vertical_reference`

#         horizontal_reference (int): Horizontal reference, see
#          :py:attr:`~geost.base.PointDataCollection.horizontal_reference`

#         header (pd.DataFrame): Header used for construction. see
#          :py:attr:`~geost.base.PointDataCollection.header`
#     """

#     def __init__(
#         self,
#         data: pd.DataFrame,
#         vertical_reference: str = "NAP",
#         horizontal_reference: int = 28992,
#         header: Optional[pd.DataFrame] = None,
#         is_inclined: bool = False,
#     ):
#         super().__init__(
#             data,
#             vertical_reference,
#             horizontal_reference,
#             header=header,
#             is_inclined=is_inclined,
#         )

#     def add_ic(
#         self,
#     ):  # Move to cpt analysis functions, use something like 'apply' function in classes
#         """
#         Calculate soil behaviour type index (Ic) for all CPT's in the collection.

#         The data is added to :py:attr:`~geost.base.PointDataCollection.header`.
#         """
#         self.data["ic"] = calc_ic(self.data["qc"], self.data["friction_number"])

#     def add_lithology(
#         self,
#     ):  # Move to cpt analysis functions, use something like 'apply' function in classes
#         """
#         Interpret lithoclass for all CPT's in the collection.

#         The data is added to :py:attr:`~geost.base.PointDataCollection.header`.
#         """
#         if "ic" not in self.data.columns:
#             self.add_ic()
#         self.data["lith"] = calc_lithology(
#             self.data["ic"], self.data["qc"], self.data["friction_number"]
#         )

#     def as_boreholecollection(self):  # No change
#         """
#         Export CptCollection to BoreholeCollection. Requires the "lith" column to be
#         present. Use the method :py:meth:`~geost.borehole.CptCollection.add_lithology`

#         Returns
#         -------
#         Instance of :class:`~geost.borehole.BoreholeCollection`
#         """
#         if "lith" not in self.data.columns:
#             raise IndexError(
#                 r"The column \"lith\" is required to convert to BoreholeCollection"
#             )

#         borehole_converted_dataframe = self.data[
#             ["nr", "x", "y", "surface", "end", "top", "bottom", "lith"]
#         ]
#         cptcollection_as_bhcollection = BoreholeCollection(
#             borehole_converted_dataframe,
#             vertical_reference=self.vertical_reference,
#             header=self.header,
#         )
#         return cptcollection_as_bhcollection



from geost import data
from geost.read import (
    get_bro_objects_from_bbox,
    get_bro_objects_from_geometry,
    read_borehole_table,
    read_collection_geopackage,
    read_cpt_table,
    read_gef_cpts,
    read_nlog_cores,
    read_pickle,
    read_uullg_tables,
    read_xml_boris,
)

# read_gef_cores,
# read_xml_cpts,
# read_xml_geological_cores,
# read_xml_geotechnical_cores,
# read_xml_soil_cores,
from geost.utils import csv_to_parquet, excel_to_parquet

__version__ = "0.2.5"



from pathlib import Path

import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal
from pyvista import MultiBlock
from shapely import get_coordinates

from geost.base import (
    BoreholeCollection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.export import geodataclass


class TestLayeredData:
    @pytest.mark.unittest
    def test_datatype(self, borehole_data):
        assert borehole_data.datatype == "layered"

    @pytest.mark.unittest
    def test_to_header(self, borehole_data):
        expected_columns = ["nr", "x", "y", "surface", "end", "geometry"]

        header = borehole_data.to_header()

        assert isinstance(header, PointHeader)
        assert_array_equal(header.gdf.columns, expected_columns)
        assert len(header.gdf) == 5
        assert header["nr"].nunique() == 5
        assert header.horizontal_reference == 28992
        assert header.vertical_reference == 5709

    @pytest.mark.unittest
    def test_to_collection(self, borehole_data):
        collection = borehole_data.to_collection()
        assert isinstance(collection, BoreholeCollection)
        assert isinstance(collection.header, PointHeader)
        assert len(collection.header) == 5

    @pytest.mark.unittest
    def test_select_by_values(self, borehole_data):
        selected = borehole_data.select_by_values("lith", ["V", "K"], how="or")
        assert isinstance(selected, LayeredData)

        expected_nrs = ["A", "B", "C", "D"]
        selected_nrs = selected["nr"].unique()
        assert_array_equal(selected_nrs, expected_nrs)

        selected = borehole_data.select_by_values("lith", ["V", "K"], how="and")

        expected_nrs = ["B", "D"]
        selected_nrs = selected["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

    @pytest.mark.unittest
    def test_slice_by_values(self, borehole_data):
        sliced = borehole_data.slice_by_values("lith", "Z")
        assert isinstance(sliced, LayeredData)

        expected_boreholes_with_sand = ["A", "C", "D", "E"]
        expected_length = 10

        assert_array_equal(sliced["nr"].unique(), expected_boreholes_with_sand)
        assert np.all(sliced["lith"] == "Z")
        assert len(sliced) == expected_length

        sliced = borehole_data.slice_by_values("lith", "Z", invert=True)

        expected_boreholes_without_sand = ["A", "B", "C", "D"]
        expected_length = 15

        assert_array_equal(sliced["nr"].unique(), expected_boreholes_without_sand)
        assert ~np.any(sliced["lith"] == "Z")
        assert len(sliced) == expected_length

    @pytest.mark.unittest
    def test_get_cumulative_thickness(self, borehole_data):
        result = borehole_data.get_cumulative_thickness("lith", "V")
        expected_boreholes_returned = ["B", "D"]
        expected_thickness = [1.9, 1.4]

        assert len(result) == 2
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_thickness)

        result = borehole_data.get_cumulative_thickness("lith", ["Z", "K"])
        expected_boreholes_returned = ["A", "B", "C", "D", "E"]
        expected_sand_thickness = [2.2, np.nan, 2.6, 0.5, 3.0]
        expected_clay_thickness = [2.0, 2.0, 2.9, 1.1, np.nan]

        assert result.shape == (5, 2)
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["K"], expected_clay_thickness)
        assert_array_almost_equal(result["Z"], expected_sand_thickness)

    @pytest.mark.unittest
    def test_get_layer_top(self, borehole_data):
        result = borehole_data.get_layer_top("lith", "V")
        expected_boreholes_returned = ["B", "D"]
        expected_tops = [1.2, 0.5]

        assert len(result) == 2
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["V"], expected_tops)

        result = borehole_data.get_layer_top("lith", ["Z", "K"])
        expected_boreholes_returned = ["A", "B", "C", "D", "E"]
        expected_sand_top = [1.5, np.nan, 2.9, 2.5, 0.0]
        expected_clay_top = [0.0, 0.0, 0.0, 0.0, np.nan]

        assert result.shape == (5, 2)
        assert_array_equal(result.index, expected_boreholes_returned)
        assert_array_almost_equal(result["Z"], expected_sand_top)
        assert_array_almost_equal(result["K"], expected_clay_top)

    @pytest.mark.unittest
    def test_slice_depth_interval(self, borehole_data):
        # Test slicing with respect to depth below the surface.
        upper, lower = 0.6, 2.4
        sliced = borehole_data.slice_depth_interval(upper, lower)
        assert isinstance(sliced, LayeredData)

        layers_per_borehole = sliced["nr"].value_counts()
        expected_layer_count = [3, 3, 3, 3, 2]

        assert len(sliced) == 14
        assert sliced["top"].min() == upper
        assert sliced["bottom"].max() == lower
        assert_array_equal(layers_per_borehole, expected_layer_count)

        # Test slicing without updating layer boundaries.
        sliced = borehole_data.slice_depth_interval(
            upper, lower, update_layer_boundaries=False
        )

        expected_tops_of_slice = [0.0, 0.6, 0.0, 0.5, 0.5]
        expected_bottoms_of_slice = [2.5, 2.5, 2.9, 2.5, 2.5]

        tops_of_slice = sliced.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()

        assert len(sliced) == 14
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slicing with respect to a vertical reference plane.
        nap_upper, nap_lower = -2, -3
        sliced = borehole_data.slice_depth_interval(
            nap_upper, nap_lower, relative_to_vertical_reference=True
        )

        expected_tops_of_slice = [2.2, 2.3, 2.25, 2.1, 1.9]
        expected_bottoms_of_slice = [3.2, 3.3, 3.25, 3.0, 2.9]

        tops_of_slice = sliced.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()

        assert len(sliced) == 11
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slices that return empty objects.
        empty_slice = borehole_data.slice_depth_interval(-2, -1)
        empty_slice_nap = borehole_data.slice_depth_interval(
            3, 2, relative_to_vertical_reference=True
        )

        assert len(empty_slice) == 0
        assert len(empty_slice_nap) == 0

        # Test slicing using only an upper boundary or lower boundary.
        upper = 4
        sliced = borehole_data.slice_depth_interval(upper)

        expected_boreholes = ["A", "C"]

        assert len(sliced) == 2
        assert_array_equal(sliced["nr"], expected_boreholes)

        nap_lower = -0.5
        sliced = borehole_data.slice_depth_interval(
            lower_boundary=nap_lower, relative_to_vertical_reference=True
        )

        bottoms_of_slice = sliced.df.groupby("nr")["bottom"].max()
        expected_bottoms_of_slice = [0.7, 0.8, 0.75, 0.6, 0.4]

        assert len(sliced) == 7
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

    @pytest.mark.unittest
    def test_select_by_condition(self, borehole_data):
        selected = borehole_data.select_by_condition(borehole_data["lith"] == "V")
        assert isinstance(selected, LayeredData)

        expected_nrs = ["B", "D"]
        assert_array_equal(selected["nr"].unique(), expected_nrs)
        assert np.all(selected["lith"] == "V")
        assert len(selected) == 4

        selected = borehole_data.select_by_condition(
            borehole_data["lith"] == "V", invert=True
        )
        assert len(selected) == 21
        assert ~np.all(selected["lith"] == "V")

    @pytest.mark.unittest
    def test_to_multiblock(self, borehole_data):
        # Test normal to multiblock.
        multiblock = borehole_data.to_multiblock("lith")
        expected_bounds = (0.0, 5.0, 0.0, 6.0, -5.25, 0.3)
        assert isinstance(multiblock, MultiBlock)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds
        assert multiblock[0].n_arrays == 2
        assert multiblock[0].n_cells == 22
        assert multiblock[0].n_points == 160

        # Test with vertical exageration.
        multiblock = borehole_data.to_multiblock("lith", vertical_factor=10)
        expected_bounds = (0.0, 5.0, 0.0, 6.0, -52.5, 3.0)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

        # Test to multiblock with respect to depth below the surface.
        multiblock = borehole_data.to_multiblock(
            "lith", relative_to_vertical_reference=False
        )
        expected_bounds = (0.0, 5.0, 0.0, 6.0, 0.0, 5.5)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

        # Test with both options.
        multiblock = borehole_data.to_multiblock(
            "lith", vertical_factor=10, relative_to_vertical_reference=False
        )
        expected_bounds = (0.0, 5.0, 0.0, 6.0, 0.0, 55.0)
        assert multiblock.n_blocks == 5
        assert multiblock.bounds == expected_bounds

    @pytest.mark.unittest
    def test_to_datafusiontools(self, borehole_data):
        # Test normal export.
        dft = borehole_data.to_datafusiontools("lith")

        expected_independent_value = [-0.2, -0.95, -1.8, -2.9, -3.75]
        expected_number_of_variables = 1

        assert len(dft) == 5
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])
        assert np.all([len(d.variables) == expected_number_of_variables for d in dft])
        assert_array_almost_equal(
            dft[0].independent_variable.value, expected_independent_value
        )

        # Test with label encoding.
        dft = borehole_data.to_datafusiontools("lith", encode=True)
        expected_number_of_variables = 3
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])
        assert np.all([len(d.variables) == expected_number_of_variables for d in dft])

        # Test without updating layer depths to NAP
        dft = borehole_data.to_datafusiontools(
            "lith", relative_to_vertical_reference=False
        )
        expected_independent_value = [0.4, 1.15, 2.0, 3.1, 3.95]

        assert_array_almost_equal(
            dft[0].independent_variable.value, expected_independent_value
        )

    @pytest.mark.unittest
    def test_to_vtm_with_file(self, borehole_data):
        outfile = Path("temp.vtm")
        outfolder = outfile.parent / r"temp"
        borehole_data.to_vtm(outfile, "lith")
        assert outfile.is_file()
        outfile.unlink()
        for f in outfolder.glob("*.vtp"):
            f.unlink()
        outfolder.rmdir()
        pass

    @pytest.mark.unittest
    def test_to_datafusiontools_with_file(self, borehole_data):
        outfile = Path("dft.pkl")
        borehole_data.to_datafusiontools("lith", outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_qgis3d(self, borehole_data):
        outfile = Path("temp.gpkg")
        borehole_data.to_qgis3d(outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_kingdom(self, borehole_data):
        outfile = Path("temp_kingdom.csv")
        tdfile = Path(outfile.parent, f"{outfile.stem}_TDCHART{outfile.suffix}")
        borehole_data.to_kingdom(outfile)
        assert outfile.is_file()
        assert tdfile.is_file()
        out_layerdata = pd.read_csv(outfile)
        out_tddata = pd.read_csv(tdfile)
        assert_array_almost_equal(
            out_layerdata["Start depth"][:6],
            np.array([0.0, 0.8, 1.5, 2.5, 3.7, 0.0]),
        )
        assert_array_almost_equal(
            out_layerdata["End depth"][:6],
            np.array([0.8, 1.5, 2.5, 3.7, 4.2, 0.6]),
        )
        assert_array_almost_equal(
            out_layerdata["Total depth"][:6],
            np.array([4.2, 4.2, 4.2, 4.2, 4.2, 3.9]),
        )
        assert_array_almost_equal(
            out_tddata["MD"],
            np.array([0, 1, 0, 1, 0, 1, 0, 1, 0, 1]),
        )
        assert_array_almost_equal(
            out_tddata["TWT"],
            np.array(
                [
                    -0.26666667,
                    0.98333333,
                    -0.4,
                    0.85,
                    -0.33333333,
                    0.91666667,
                    -0.13333333,
                    1.11666667,
                    0.13333333,
                    1.38333333,
                ]
            ),
        )
        outfile.unlink()
        tdfile.unlink()

    @pytest.mark.unittest
    def test_create_geodataframe_3d(self, borehole_data):
        relative_to_vertical_reference = True
        gdf = borehole_data._create_geodataframe_3d(
            relative_to_vertical_reference, crs=28992
        )

        first_line_coords = get_coordinates(gdf["geometry"].iloc[0], include_z=True)
        expected_coords = [[2.0, 3.0, 0.21], [2.0, 3.0, -0.59]]

        assert all(gdf.geom_type == "LineString")
        assert_array_almost_equal(first_line_coords, expected_coords)

        relative_to_vertical_reference = False
        gdf = borehole_data._create_geodataframe_3d(
            relative_to_vertical_reference, crs=28992
        )

        first_line_coords = get_coordinates(gdf["geometry"].iloc[0], include_z=True)
        expected_coords = [[2.0, 3.0, 0.01], [2.0, 3.0, 0.81]]

        assert all(gdf.geom_type == "LineString")
        assert_array_almost_equal(first_line_coords, expected_coords)

    @pytest.mark.unittest
    def test_change_depth_values(self, borehole_data):
        borehole = borehole_data.select_by_values("nr", "A").df
        borehole = borehole_data._change_depth_values(borehole)

        expected_top = [0.2, -0.6, -1.3, -2.3, -3.5]
        expected_bottom = [-0.6, -1.3, -2.3, -3.5, -4.0]

        assert_array_almost_equal(borehole["top"], expected_top)
        assert_array_almost_equal(borehole["bottom"], expected_bottom)

    @pytest.mark.unittest
    def test_check_correct_instance(self, borehole_data):
        inst = "string"
        inst = borehole_data._check_correct_instance(inst)
        assert isinstance(inst, list)

        inst = ["list of strings"]
        inst = borehole_data._check_correct_instance(inst)
        assert isinstance(inst, list)

    @pytest.mark.unittest
    def test_to_csv_mixin(self, borehole_data):
        outfile = Path("temp.csv")
        borehole_data.to_csv(outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_parquet_mixin(self, borehole_data):
        outfile = Path("temp.parquet")
        borehole_data.to_parquet(outfile)
        assert outfile.is_file()
        outfile.unlink()


class TestDiscreteData:
    @pytest.mark.unittest
    def test_datatype(self, cpt_data):
        assert cpt_data.datatype == "discrete"

    @pytest.mark.unittest
    def test_to_header(self, cpt_data):
        header = cpt_data.to_header()
        expected_columns = ["nr", "x", "y", "surface", "end", "geometry"]

        header = cpt_data.to_header()

        assert isinstance(header, PointHeader)
        assert_array_equal(header.gdf.columns, expected_columns)
        assert len(header.gdf) == 2
        assert header["nr"].nunique() == 2
        assert header.horizontal_reference == 28992
        assert header.vertical_reference == 5709

    @pytest.mark.unittest
    def test_to_collection(self, cpt_data):
        collection = cpt_data.to_collection()
        assert isinstance(collection, CptCollection)
        assert isinstance(collection.header, PointHeader)
        assert len(collection.header) == 2

    @pytest.mark.unittest
    def test_select_by_values(self, cpt_data):
        selected = cpt_data.select_by_values("nr", "a")
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 10
        assert_array_equal(selected["nr"].unique(), "a")

    @pytest.mark.unittest
    def test_select_by_condition(self, cpt_data):
        selected = cpt_data.select_by_condition(
            (cpt_data["qc"] > 0.3) & (cpt_data["qc"] < 0.4)
        )
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 5
        assert np.all((selected["qc"] > 0.3) & (selected["qc"] < 0.4))
        assert_array_equal(selected.df.index, [2, 3, 4, 5, 6])

        selected = cpt_data.select_by_condition(
            (cpt_data["qc"] < 0.35) & (cpt_data["fs"] < 0.2)
        )
        assert len(selected) == 3
        assert np.all(selected["qc"] < 0.35)
        assert np.all(selected["fs"] < 0.2)
        assert_array_equal(selected.df.index, [0, 1, 2])

    @pytest.mark.unittest
    def test_slice_depth_interval(self, cpt_data):
        # Selection with respect to surface level
        selected = cpt_data.slice_depth_interval(2, 3)
        assert isinstance(selected, DiscreteData)
        assert len(selected) == 4
        assert_array_equal(selected["depth"], [2, 3, 2, 3])

        # Selection with respect to vertical reference plane
        selected = cpt_data.slice_depth_interval(
            1.9, 0.9, relative_to_vertical_reference=True
        )
        assert len(selected) == 1
        assert_array_equal(selected["depth"], [1])
        assert_array_equal(selected["nr"], ["a"])

        # Selection with respect to surface level using one limit
        selected = cpt_data.slice_depth_interval(lower_boundary=0.9)
        assert len(selected) == 2
        assert_array_equal(selected["depth"], [0, 0])
        assert_array_equal(selected["nr"], ["a", "b"])

        # Selection with respect to vertical reference plane using one limit
        selected = cpt_data.slice_depth_interval(
            lower_boundary=0.9, relative_to_vertical_reference=True
        )
        assert len(selected) == 2
        assert_array_equal(selected["depth"], [0, 1])
        assert np.all(selected["nr"] == "a")



from pathlib import Path
from typing import Iterable

import geopandas as gpd
import numpy as np
import pandas as pd
import rioxarray
import xarray as xr

from geost.utils import inform_user, warn_user

warn = warn_user(lambda warning_info: print(warning_info))
inform = inform_user(lambda info: print(info))


def check_gdf_instance(
    gdf_or_path: str | Path | gpd.GeoDataFrame,
) -> gpd.GeoDataFrame:
    """
    Check if the argument is already a geodataframe or pointing to a file
    containing geometries

    Parameters
    ----------
    gdf_or_path : str | Path | gpd.GeoDataFrame
        The geodataframe or file that can be read and parsed to a geodataframe

    Returns
    -------
    gpd.GeoDataFrame
        An instance of a geopandas geodataframe
    """
    if isinstance(gdf_or_path, str | Path):
        gdf_or_path = Path(gdf_or_path)
        filetype = gdf_or_path.suffix
        if filetype in (".parquet", ".geoparquet"):
            gdf = gpd.read_parquet(gdf_or_path)
        else:
            gdf = gpd.read_file(gdf_or_path)
    elif isinstance(gdf_or_path, gpd.GeoDataFrame):
        gdf = gdf_or_path

    # Make sure you don't get a view of gdf returned
    gdf = gdf.copy()

    return gdf


def check_and_coerce_crs(gdf: gpd.GeoDataFrame, to_crs: int):
    """
    Check the CRS of a geodataframe against given crs.

    If the geodataframe has no CRS, makes the geodataframe assume the given CRS. The
    user is warned when this occurs.

    If the geodataframe has a different known CRS, inform user and coerce the CRS
    to the given CRS (to_crs argument).

    Parameters
    ----------
    gdf : GeoDataFrame
        Geodataframe to be converted to the given CRS.
    to_crs : int
        EPSG number of the CRS to check and coerce to.

    Returns
    -------
    GeoDataFrame
        Geodataframe coerced to have the desired CRS
    """
    if gdf.crs is None:
        gdf.crs = to_crs
        warn(
            "The selection geometry has no crs! Assuming it is the same as the "
            + f"horizontal_reference (epsg:{to_crs}) of this "
            + "collection. PLEASE CHECK WHETHER THIS IS CORRECT!",
        )
    elif gdf.crs != to_crs:
        gdf = gdf.to_crs(to_crs)
        inform(
            "The crs of the selection geometry does not match the horizontal "
            + "reference of the collection. The selection geometry was coerced "
            + f"to epsg:{to_crs} automatically"
        )
    return gdf


def select_points_within_bbox(
    gdf: str | Path | gpd.GeoDataFrame,
    xmin: float | int,
    ymin: float | int,
    xmax: float | int,
    ymax: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on a user-given bounding box.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    xmin : float | int
        Minimum x-coordinate of the bounding box.
    ymin : float | int
        Minimum y-coordinate of the bounding box.
    xmax : float | int
        Maximum x-coordinate of the bounding box.
    ymax : float | int
        Maximum y-coordinate of the bounding box.
    invert : bool, optional
        Invert the selection, so select all objects outside of the
        bounding box in this case, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = check_gdf_instance(gdf)

    # Selection logic
    x = gdf["geometry"].x
    y = gdf["geometry"].y
    if invert:
        gdf_selected = gdf[(x < xmin) | (x > xmax) | (y < ymin) | (y > ymax)]
    else:
        gdf_selected = gdf[(x >= xmin) & (x <= xmax) & (y >= ymin) & (y <= ymax)]
    return gdf_selected


def select_points_near_points(
    gdf: str | Path | gpd.GeoDataFrame,
    point_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on point geometries and a buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    point_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int
        Buffer distance for selection geometries.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = check_gdf_instance(gdf)
    point_gdf = check_gdf_instance(point_gdf)

    # Selection logic
    data_points = np.array([gdf["geometry"].x, gdf["geometry"].y]).transpose()
    query_points = np.array(
        [point_gdf["geometry"].x, point_gdf["geometry"].y]
    ).transpose()

    bool_array = np.full(len(data_points), False)
    for query_point in query_points:
        distance = np.sqrt(
            (query_point[0] - data_points[:, 0]) ** 2
            + (query_point[1] - data_points[:, 1]) ** 2
        )
        bool_array += distance < buffer
    if invert:
        bool_array = np.invert(bool_array)
    gdf_selected = gdf[bool_array]
    return gdf_selected


def select_points_near_lines(
    gdf: str | Path | gpd.GeoDataFrame,
    line_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on line geometries and a buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    line_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int
        Buffer distance for selection geometries.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = check_gdf_instance(gdf)
    line_gdf = check_gdf_instance(line_gdf)

    # Selection logic
    line_gdf["geometry"] = line_gdf.buffer(distance=buffer)
    if invert:
        gdf_selected = gdf[~gdf.geometry.within(line_gdf.union_all())]
    else:
        gdf_selected = gdf[gdf.geometry.within(line_gdf.union_all())]
    return gdf_selected


def select_points_within_polygons(
    gdf: str | Path | gpd.GeoDataFrame,
    polygon_gdf: str | Path | gpd.GeoDataFrame,
    buffer: float | int = 0,
    invert: bool = False,
) -> gpd.GeoDataFrame:
    """
    Make a selection of point geometries based on polygon geometries and an optional
    buffer.

    Parameters
    ----------
    gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select from.
    polygon_gdf : str | Path | gpd.GeoDataFrame
        Geodataframe (or file that can be parsed to a geodataframe) to select with.
    buffer : float | int, optional
        Optional buffer distance around the polygon selection geometries, by default 0.
    invert : bool, optional
        Invert the selection, by default False.

    Returns
    -------
    gpd.GeoDataFrame
        Geodataframe containing only selected geometries.
    """
    # Instance checks and coerce to geodataframe if required
    gdf = check_gdf_instance(gdf)
    polygon_gdf = check_gdf_instance(polygon_gdf)

    # Selection logic
    if buffer > 0:
        polygon_select = polygon_gdf.copy()
        polygon_select["geometry"] = polygon_gdf.geometry.buffer(buffer)
    else:
        polygon_select = polygon_gdf

    if invert:
        gdf_selected = gdf[~gdf.geometry.within(polygon_select.union_all())]
    else:
        gdf_selected = gdf[gdf.geometry.within(polygon_select.union_all())]

    return gdf_selected


def find_area_labels(
    point_geodataframe: gpd.GeoDataFrame,
    polygon_geodataframe: gpd.GeoDataFrame,
    column_name: str | Iterable,
) -> pd.Series:
    """
    Function to find labels associated with polygon geometries for a series of queried
    point geometries. Basically a spatial join between the point and polygon dataframe.

    Parameters
    ----------
    point_geodataframe : gpd.GeoDataFrame
        Geodataframe with point geometries for which you want to find in which polygon
        geometries they are located.
    polygon_geodataframe : gpd.GeoDataFrame
        Geodataframe with polygon geometries
    column_name : str | Iterable
        Label of the polygon geometries to use for assigning to the queried points.
        Given as a string or iterable of strings in case you'd like to find multiple
        labels.

    Returns
    -------
    pandas.Series
        Series with labels from the polygon geometries for each point.
    """
    if not isinstance(column_name, str):
        column_name = list(column_name)
    joined = gpd.sjoin(point_geodataframe, polygon_geodataframe)[column_name]
    # Remove any duplicated indices, which may sometimes happen
    area_labels = joined[~joined.index.duplicated()]
    return area_labels


def get_raster_values(
    x: np.ndarray, y: np.ndarray, raster_to_read: str | Path | xr.DataArray
) -> np.ndarray:
    """
    Return sampled values from a raster at the given (x, y) locations.

    Parameters
    ----------
    x : np.ndarray
        1D array of x-coordinates, same length as 'y'.
    y : np.ndarray
        1D array of y-coordinates, same length as 'x'.
    raster_to_read : str | Path | xr.DataArray
        Location of a raster file or an xr.DataArray with dimensions 'x' and 'y'. This
        raster is used to sample values from at all (x, y) locations.

    Returns
    -------
    np.ndarray
        1D array of sampled values
    """
    if isinstance(raster_to_read, (str, Path)):
        raster_to_read = rioxarray.open_rasterio(raster_to_read).squeeze()

    if set(raster_to_read.dims) != set(("y", "x")):
        raise TypeError(
            "The xr.DataArray to sample from does not have the "
            + "required 'x' and 'y' dimensions"
        )

    xmin, ymin, xmax, ymax = raster_to_read.rio.bounds()
    outside_x = (x < xmin) | (x > xmax)
    outside_y = (y < ymin) | (y > ymax)

    surface_levels = raster_to_read.sel(
        x=xr.DataArray(x, dims=("loc")),
        y=xr.DataArray(y, dims=("loc")),
        method="nearest",
    ).values

    surface_levels[outside_x | outside_y] = np.nan

    return surface_levels



from typing import NamedTuple

from geost.validate.validate import Check, Column, numeric, stringlike


class ValidationSchemas(NamedTuple):
    headerschema_point = {
        "nr": Column(stringlike),
        "x": Column(numeric),
        "y": Column(numeric),
        "surface": Column(numeric),
        "end": Column(numeric, checks=Check("<", "surface", report_by="nr")),
    }
    dataschema_layered_point = {
        "nr": Column(stringlike),
        "x": Column(numeric),
        "y": Column(numeric),
        "surface": Column(numeric),
        "end": Column(numeric),
        "top": Column(numeric),
        "bottom": Column(numeric, checks=Check(">", "top", report_by="nr")),
    }
    dataschema_discrete_point = {
        "nr": Column(stringlike),
        "x": Column(numeric),
        "y": Column(numeric),
        "surface": Column(numeric),
        "end": Column(numeric),
        "depth": Column(numeric),
    }
    dataschema_inclined_point = {
        "x_bot": Column(numeric),
        "y_bot": Column(numeric),
    }

    # Below schemas are not yet in use
    boreholeschema = {
        "lith": Column(stringlike),
    }
    cptschema = {
        "length": Column(float),
        "qc": Column(float),
    }



from pathlib import Path

import geopandas as gpd
import numpy as np
import pandas as pd
import pytest
import rioxarray as rio
import xarray as xr
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
    assert_equal,
)
from pyvista import MultiBlock
from shapely.geometry import LineString, Point, Polygon

from geost.base import BoreholeCollection, LayeredData, PointHeader
from geost.export import geodataclass


class TestCollection:
    nlog_stratstelsel_parquet = (
        Path(__file__).parent / "data/test_nlog_stratstelsel_20230807.parquet"
    )

    @pytest.fixture
    def borehole_df_ok(self):
        nr = np.full(10, "B-01")
        x = np.full(10, 139370)
        y = np.full(10, 455540)
        mv = np.full(10, 1.0)
        end = np.full(10, -4.0)
        top = np.array([0, 0.5, 1, 1.5, 2, 3, 4, 5, 6, 7])
        bottom = np.array([0.5, 1, 1.5, 2, 3, 4, 5, 6, 7, 8])
        data_string = np.array(
            ["K", "Kz", "K", "Ks3", "Ks2", "V", "Zk", "Zs", "Z", "Z"]
        )
        data_int = np.arange(0, 10, dtype=np.int64)
        data_float = np.arange(0, 5, 0.5, dtype=np.float64)

        dataframe = pd.DataFrame(
            {
                "nr": nr,
                "x": x,
                "y": y,
                "surface": mv,
                "end": end,
                "top": top,
                "bottom": bottom,
                "data_string": data_string,
                "data_int": data_int,
                "data_float": data_float,
            }
        )

        return dataframe

    @pytest.fixture
    def borehole_df_bad_validation(self):
        nr = np.full(10, 1)
        x = np.full(10, 139370)
        y = np.full(10, 455540)
        mv = np.full(10, 1.0)
        end = np.full(10, 4.0)
        top = np.array([0, 0.5, 1, 1.5, 2, 3, 4, 5, 6, 7])
        bottom = np.array([0.5, 1, 1.5, 2, 3, 4, 5, 6, 7, 6.5])
        data_string = np.array(
            ["K", "Kz", "K", "Ks3", "Ks2", "V", "Zk", "Zs", "Z", "Z"]
        )
        data_int = np.arange(0, 10, dtype=np.int64)
        data_float = np.arange(0, 5, 0.5, dtype=np.float64)

        dataframe = pd.DataFrame(
            {
                "nr": nr,
                "x": x,
                "y": y,
                "surface": mv,
                "end": end,
                "top": top,
                "bottom": bottom,
                "data_string": data_string,
                "data_int": data_int,
                "data_float": data_float,
            }
        )

        return dataframe

    @pytest.fixture
    def header_missing_object(self):
        dataframe = pd.DataFrame(
            {
                "nr": ["B-02"],
                "x": [100000],
                "y": [400000],
                "surface": [0],
                "end": [-8],
            }
        )

        return dataframe

    @pytest.fixture
    def header_surplus_objects(self):
        dataframe = pd.DataFrame(
            {
                "nr": ["B-01", "B-02", "B-03"],
                "x": [139370, 100000, 110000],
                "y": [455540, 400000, 410000],
                "surface": [1, 0, -1],
                "end": [-4, -8, -9],
            }
        )

        return dataframe

    @pytest.fixture
    def update_raster(self):
        x_coors = np.arange(127000, 128500, 500)
        y_coors = np.arange(503000, 501500, -500)
        data = np.ones((3, 3))
        array = xr.DataArray(data, {"x": x_coors, "y": y_coors})
        return array

    @pytest.mark.unittest
    def test_get_single_object(self, borehole_collection):
        borehole_collection_single_selection = borehole_collection.get("A")
        assert borehole_collection_single_selection.header.gdf.iloc[0, 0] == "A"

    @pytest.mark.unittest
    def test_get_multiple_objects(self, borehole_collection):
        borehole_collection_multi_selection = borehole_collection.get(["A", "B"])
        assert list(borehole_collection_multi_selection.header["nr"].unique()) == [
            "A",
            "B",
        ]

    @pytest.mark.unittest
    def test_add_header_column_to_data(self, borehole_collection):
        borehole_collection.header["test_data"] = [
            i for i in range(len(borehole_collection.header))
        ]
        borehole_collection.add_header_column_to_data("test_data")

        assert_allclose(borehole_collection.get("A").data["test_data"], 0)
        assert_allclose(borehole_collection.get("B").data["test_data"], 1)

    @pytest.mark.unittest
    def test_change_vertical_reference(self, borehole_collection):
        assert borehole_collection.vertical_reference == 5709
        borehole_collection.change_vertical_reference("Ostend height")
        assert borehole_collection.vertical_reference == 5710

    @pytest.mark.unittest
    def test_change_horizontal_reference(self, borehole_collection):
        assert borehole_collection.horizontal_reference == 28992
        borehole_collection.change_horizontal_reference(32631)
        assert borehole_collection.horizontal_reference == 32631
        assert_almost_equal(
            borehole_collection.data["x"][0], borehole_collection.header["x"][0]
        )
        assert_almost_equal(borehole_collection.data["x"][0], 523403.281803, decimal=5)
        assert_almost_equal(
            borehole_collection.data["y"][0], borehole_collection.header["y"][0]
        )
        assert_almost_equal(borehole_collection.data["y"][0], 5313546.187669, decimal=5)

    @pytest.mark.unittest
    def test_change_horizontal_reference_also_data_columns_inclined(
        self, nlog_borehole_collection
    ):
        assert nlog_borehole_collection.horizontal_reference == 28992
        nlog_borehole_collection.change_horizontal_reference(32631)
        assert nlog_borehole_collection.horizontal_reference == 32631
        assert_almost_equal(
            nlog_borehole_collection.header["x"][0], 532641.76, decimal=2
        )
        assert_almost_equal(
            nlog_borehole_collection.header["y"][0], 6.170701e6, decimal=0
        )
        assert_almost_equal(
            nlog_borehole_collection.data["x_bot"][0], 532650.77, decimal=2
        )
        assert_almost_equal(
            nlog_borehole_collection.data["y_bot"][0], 6.170700e6, decimal=0
        )

    @pytest.mark.unittest
    def test_select_within_bbox(self, borehole_collection):
        borehole_collection_selected = borehole_collection.select_within_bbox(
            1.5, 1.5, 3.5, 5
        )
        # The selection results boreholes 'A' and 'D'
        assert all(borehole_collection_selected.header.gdf["nr"].unique() == ["A", "D"])
        assert all(borehole_collection_selected.data.df["nr"].unique() == ["A", "D"])

    @pytest.mark.unittest
    def test_select_with_points(self, borehole_collection):
        selection_points = [Point(1, 1), Point(4, 4), Point(1, 4), Point(4, 1)]
        selection_gdf = gpd.GeoDataFrame(
            {"id": [0, 1, 2, 3]}, geometry=selection_points
        )
        collection_sel = borehole_collection.select_with_points(selection_gdf, 1.1)
        collection_sel_inverted = borehole_collection.select_with_points(
            selection_gdf, 1.1, invert=True
        )
        assert collection_sel.n_points == 3
        assert collection_sel_inverted.n_points == 2

    @pytest.mark.unittest
    def test_select_with_lines(self, borehole_collection):
        selection_lines = [LineString([[1, 1], [5, 5]]), LineString([[1, 5], [5, 1]])]
        selection_gdf = gpd.GeoDataFrame({"id": [0, 1]}, geometry=selection_lines)
        collection_sel = borehole_collection.select_with_lines(selection_gdf, 0.6)
        collection_sel_inverted = borehole_collection.select_with_lines(
            selection_gdf, 0.6, invert=True
        )
        assert collection_sel.n_points == 2
        assert collection_sel_inverted.n_points == 3

    @pytest.mark.unittest
    def test_select_within_polygons(self, borehole_collection):
        selection_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 3)))]
        selection_gdf = gpd.GeoDataFrame({"id": [0]}, geometry=selection_polygon)

        # Polygon selection without buffer
        collection_sel = borehole_collection.select_within_polygons(selection_gdf)
        collection_sel_inverted = borehole_collection.select_within_polygons(
            selection_gdf, invert=True
        )
        assert collection_sel.n_points == 1
        assert collection_sel_inverted.n_points == 4

        # Polygon selection with a buffer
        collection_sel_buff = borehole_collection.select_within_polygons(
            selection_gdf, buffer=0.7
        )
        collection_sel_inverted_buff = borehole_collection.select_within_polygons(
            selection_gdf, buffer=0.7, invert=True
        )
        assert collection_sel_buff.n_points == 2
        assert collection_sel_inverted_buff.n_points == 3

    @pytest.mark.unittest
    def test_select_by_depth(self, borehole_collection):
        assert borehole_collection.select_by_depth(top_min=0).n_points == 4
        assert borehole_collection.select_by_depth(top_max=0).n_points == 1
        assert borehole_collection.select_by_depth(end_min=-3.5).n_points == 2
        assert (
            borehole_collection.select_by_depth(
                top_min=0, top_max=0.3, end_max=-4
            ).n_points
            == 2
        )

    @pytest.mark.unittest
    def test_select_by_length(self, borehole_collection):
        assert borehole_collection.select_by_length(min_length=4).n_points == 2
        assert borehole_collection.select_by_length(max_length=4).n_points == 3
        assert (
            borehole_collection.select_by_length(min_length=3, max_length=5).n_points
            == 4
        )

    @pytest.mark.unittest
    def test_select_by_values(self, borehole_collection):
        selected = borehole_collection.select_by_values("lith", ["V", "K"], how="or")

        expected_nrs = ["A", "B", "C", "D"]
        selected_nrs = selected.data["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

        selected = borehole_collection.select_by_values("lith", ["V", "K"], how="and")

        expected_nrs = ["B", "D"]
        selected_nrs = selected.data["nr"].unique()

        assert_array_equal(selected_nrs, expected_nrs)

    @pytest.mark.unittest
    def test_slice_depth_interval(self, borehole_collection):
        # Test slicing with respect to depth below the surface.
        upper, lower = 0.6, 2.4
        sliced = borehole_collection.slice_depth_interval(upper, lower)

        layers_per_borehole = sliced.data["nr"].value_counts()
        expected_layer_count = [3, 3, 3, 3, 2]

        assert len(sliced.data) == 14
        assert sliced.n_points == 5
        assert sliced.data["top"].min() == upper
        assert sliced.data["bottom"].max() == lower
        assert_array_equal(layers_per_borehole, expected_layer_count)

        # Test slicing without updating layer boundaries.
        sliced = borehole_collection.slice_depth_interval(
            upper, lower, update_layer_boundaries=False
        )

        expected_tops_of_slice = [0.0, 0.6, 0.0, 0.5, 0.5]
        expected_bottoms_of_slice = [2.5, 2.5, 2.9, 2.5, 2.5]

        tops_of_slice = sliced.data.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()

        assert len(sliced.data) == 14
        assert sliced.n_points == 5
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slicing with respect to a vertical reference plane.
        nap_upper, nap_lower = -2, -3
        sliced = borehole_collection.slice_depth_interval(
            nap_upper, nap_lower, relative_to_vertical_reference=True
        )

        expected_tops_of_slice = [2.2, 2.3, 2.25, 2.1, 1.9]
        expected_bottoms_of_slice = [3.2, 3.3, 3.25, 3.0, 2.9]

        tops_of_slice = sliced.data.df.groupby("nr")["top"].min()
        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()

        assert len(sliced.data) == 11
        assert sliced.n_points == 5
        assert_array_equal(tops_of_slice, expected_tops_of_slice)
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

        # Test slices that return empty objects.
        empty_slice = borehole_collection.slice_depth_interval(-2, -1)
        empty_slice_nap = borehole_collection.slice_depth_interval(
            3, 2, relative_to_vertical_reference=True
        )

        assert len(empty_slice.data) == 0
        assert len(empty_slice_nap.data) == 0
        assert empty_slice.n_points == 0
        assert empty_slice_nap.n_points == 0

        # Test slicing using only an upper boundary or lower boundary.
        upper = 4
        sliced = borehole_collection.slice_depth_interval(upper)

        expected_boreholes = ["A", "C"]

        assert len(sliced.data) == 2
        assert sliced.n_points == 2
        assert_array_equal(sliced.data["nr"], expected_boreholes)

        nap_lower = -0.5
        sliced = borehole_collection.slice_depth_interval(
            lower_boundary=nap_lower, relative_to_vertical_reference=True
        )

        bottoms_of_slice = sliced.data.df.groupby("nr")["bottom"].max()
        expected_bottoms_of_slice = [0.7, 0.8, 0.75, 0.6, 0.4]

        assert len(sliced.data) == 7
        assert sliced.n_points == 5
        assert_array_equal(bottoms_of_slice, expected_bottoms_of_slice)

    @pytest.mark.unittest
    def test_slice_by_values(self, borehole_collection):
        sliced = borehole_collection.slice_by_values("lith", "Z")

        expected_boreholes_with_sand = ["A", "C", "D", "E"]
        expected_length = 10

        assert_array_equal(sliced.data["nr"].unique(), expected_boreholes_with_sand)
        assert np.all(sliced.data["lith"] == "Z")
        assert len(sliced.data) == expected_length
        assert sliced.n_points == 4

        sliced = borehole_collection.slice_by_values("lith", "Z", invert=True)

        expected_boreholes_without_sand = ["A", "B", "C", "D"]
        expected_length = 15

        assert_array_equal(sliced.data["nr"].unique(), expected_boreholes_without_sand)
        assert ~np.any(sliced.data["lith"] == "Z")
        assert len(sliced.data) == expected_length
        assert sliced.n_points == 4

    @pytest.mark.unittest
    def test_select_by_condition(self, borehole_collection):
        # same selection with this method is used in test_data_objects
        selected = borehole_collection.select_by_condition(
            borehole_collection.data["lith"] == "V"
        )
        assert isinstance(selected, BoreholeCollection)
        assert selected.n_points == 2

    @pytest.mark.integrationtest
    def test_validation_pass(self, capfd, borehole_df_ok):
        LayeredData(borehole_df_ok).to_collection()
        out, err = capfd.readouterr()
        # Since no warning line was printed, the length of out must be 0
        assert_equal(len(out), 0)

    @pytest.mark.integrationtest
    def test_validation_fail(self, capfd, borehole_df_bad_validation):
        LayeredData(borehole_df_bad_validation).to_collection()
        out, err = capfd.readouterr()
        # Check if required warnings are printed. Note that changing warning messages
        # will make this test fail.
        assert 'but is required to be "stringlike type"' in out
        assert 'data in column "bottom" failed check "> top" for 1 rows: [1]' in out
        assert 'data in column "end" failed check "< surface" for 1 rows: [1]' in out

    @pytest.mark.integrationtest
    def test_header_mismatch(
        self, capfd, borehole_df_ok, header_missing_object, header_surplus_objects
    ):
        # Situation #1: More unique objects in data table than listed in header
        BoreholeCollection(
            PointHeader(header_missing_object, 28992), LayeredData(borehole_df_ok)
        )
        out, err = capfd.readouterr()
        assert "Header does not cover all unique objects in data" in out

        # Situation #2: More objects in header table than in data table
        BoreholeCollection(
            PointHeader(header_surplus_objects, 28992), LayeredData(borehole_df_ok)
        )
        out, err = capfd.readouterr()
        assert "Header covers more/other objects than present in the data table" in out

    @pytest.mark.unittest
    def test_get_area_labels(self, borehole_collection):
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 4)))]
        label_gdf = gpd.GeoDataFrame({"id": [1]}, geometry=label_polygon)
        # Return variant
        output = borehole_collection.get_area_labels(label_gdf, "id")
        assert_almost_equal(output["id"].sum(), 2)
        # In-place variant
        borehole_collection.get_area_labels(label_gdf, "id", include_in_header=True)
        assert_almost_equal(borehole_collection.header["id"].sum(), 2)

    @pytest.mark.unittest
    def test_get_area_labels_multiple(self, borehole_collection):
        label_polygon = [Polygon(((2, 1), (5, 4), (4, 5), (1, 4)))]
        label_gdf = gpd.GeoDataFrame(
            {"id": [1], "col2": ["string_data"]}, geometry=label_polygon
        )
        # Return variant
        output = borehole_collection.get_area_labels(label_gdf, ["id", "col2"])
        assert_almost_equal(output["id"].sum(), 2)
        assert_equal(output["col2"].value_counts()["string_data"], 2)
        # In-place variant
        borehole_collection.get_area_labels(
            label_gdf, ("id", "col2"), include_in_header=True
        )
        assert_almost_equal(borehole_collection.header["id"].sum(), 2)
        assert_equal(
            borehole_collection.header["col2"].value_counts()["string_data"], 2
        )

    @pytest.mark.unittest
    def test_to_multiblock(self, borehole_collection):
        # More detailed tests are in TestLayeredData in test_data_objects.py
        multiblock = borehole_collection.to_multiblock("lith")
        assert isinstance(multiblock, MultiBlock)

    @pytest.mark.unittest
    def test_to_vtm(self, borehole_collection):
        outfile = Path("temp.vtm")
        outfolder = outfile.parent / r"temp"
        borehole_collection.to_vtm(outfile, "lith")
        assert outfile.is_file()
        outfile.unlink()
        for f in outfolder.glob("*.vtp"):
            f.unlink()
        outfolder.rmdir()

    @pytest.mark.unittest
    def test_to_datafusiontools(self, borehole_collection):
        # More detailed tests are in TestLayeredData in test_data_objects.py
        dft = borehole_collection.to_datafusiontools("lith")
        assert np.all([isinstance(d, geodataclass.Data) for d in dft])

        outfile = Path("dft.pkl")
        borehole_collection.to_datafusiontools("lith", outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_qgis3d(self, borehole_collection):
        outfile = Path("temp.gpkg")
        borehole_collection.to_qgis3d(outfile)
        assert outfile.is_file()
        outfile.unlink()

    @pytest.mark.unittest
    def test_to_geoparquet(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.geoparquet"
        borehole_collection.to_geoparquet(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_shape(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.shp"
        borehole_collection.to_shape(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_geopackage(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_geopackage(outfile)
        assert outfile.is_file()
        layers = gpd.list_layers(outfile)
        assert_array_equal(layers["name"], ["header", "data"])
        assert_array_equal(layers["geometry_type"], ["Point", None])

    @pytest.mark.unittest
    def test_to_parquet(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_parquet(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_csv(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.gpkg"
        borehole_collection.to_csv(outfile)
        assert outfile.is_file()

    @pytest.mark.unittest
    def test_to_pickle(self, borehole_collection, tmp_path):
        outfile = tmp_path / r"test_export.pkl"
        borehole_collection.to_pickle(outfile)
        assert outfile.is_file()

    # @pytest.mark.integrationtest
    # def test_surface_level_update(self, borehole_collection, update_raster):
    #     borehole_collection.update_surface_level_from_raster(update_raster, how="replace")
    #     print("stop")


class TestBoreholeCollection:
    @pytest.mark.unittest
    def test_get_cumulative_thickness_multiple(self, borehole_collection):
        borehole_collection.get_cumulative_thickness(
            "lith", ["Z", "K"], include_in_header=True
        )
        expected_clay_thickness = [2.0, 2.0, 2.9, 1.1, 0.0]
        expected_sand_thickness = [2.2, 0.0, 2.6, 0.5, 3.0]

        assert_almost_equal(
            borehole_collection.header["K_thickness"], expected_clay_thickness
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

        # Single query
        borehole_collection.get_cumulative_thickness(
            "lith", "Z", include_in_header=True
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

    @pytest.mark.unittest
    def test_get_cumulative_thickness_single(self, borehole_collection):
        expected_sand_thickness = [2.2, 0.0, 2.6, 0.5, 3.0]

        # Single query
        borehole_collection.get_cumulative_thickness(
            "lith", "Z", include_in_header=True
        )
        assert_almost_equal(
            borehole_collection.header["Z_thickness"], expected_sand_thickness
        )

    @pytest.mark.unittest
    def test_get_layer_top(self, borehole_collection):
        borehole_collection.get_layer_top("lith", ["Z", "K"], include_in_header=True)

        expected_sand_top = [1.5, np.nan, 2.9, 2.5, 0.0]
        expected_clay_top = [0.0, 0.0, 0.0, 0.0, np.nan]

        assert_almost_equal(borehole_collection.header["K_top"], expected_clay_top)
        assert_almost_equal(borehole_collection.header["Z_top"], expected_sand_top)

    @pytest.mark.unittest
    def test_to_kingdom(self, borehole_collection):
        outfile = Path("temp_kingdom.csv")
        tdfile = Path(outfile.parent, f"{outfile.stem}_TDCHART{outfile.suffix}")
        borehole_collection.to_kingdom(outfile)
        assert outfile.is_file()
        assert tdfile.is_file()
        outfile.unlink()
        tdfile.unlink()


class TestCptCollection:
    @pytest.mark.unittest
    def test_slice_depth_interval(self, cpt_collection):
        upper, lower = 0.6, 4.4
        sliced = cpt_collection.slice_depth_interval(upper, lower)

        assert len(sliced.data) == 8
        assert sliced.n_points == 2
        assert sliced.data["depth"].min() >= upper
        assert sliced.data["depth"].max() <= lower

        upper, lower = 1.9, 0.9  # Elevations in NAP
        sliced = cpt_collection.slice_depth_interval(
            upper, lower, relative_to_vertical_reference=True
        )
        assert sliced.n_points == 1
        assert len(sliced.data) == 1



import re
from pathlib import Path
from typing import NamedTuple, Union

import numpy as np
import pandas as pd
from lxml import etree

from geost.io.parsers.parser_utils import DDCoord, RDCoord, safe_coerce


class LayerSoilCore(NamedTuple):
    top: Union[int, float]
    bottom: Union[int, float]
    horizon: str
    standard_name: str
    pedologic_name: str
    # TODO: not all the below are always in the xml
    # om: str
    # ca: str
    # gravel: str
    # shells: str
    # clay_perc: Union[int, float]


class SoilCore:
    def __init__(self, xml: str | Path | etree._Element):
        if isinstance(xml, (str, Path)):
            self.root = etree.parse(xml).getroot()
        elif isinstance(xml, etree._Element):
            self.root = xml
        else:
            raise TypeError(
                "Input datatype for reader not allowed. Must be an lxml Element "
                "tree or path to an xml file."
            )

        self.broid = None
        self.x = None
        self.y = None
        self.z = None
        self.enddepth = None
        self.quality = None
        self.crs = None
        self.reference_level = None

        self.__set_namespaces()
        self.__get_data()
        self._set_header_info()

        self.df = pd.DataFrame(self.parse_layers())
        attr_dict = {
            "nr": "broid",
            "x": "x",
            "y": "y",
            "surface": "z",
            "end": "enddepth",
            "codegroup": "codegroup",
            "soilclass": "soilclass",
        }
        for attr in ["nr", "x", "y", "surface", "end", "codegroup", "soilclass"][::-1]:
            self.df.insert(0, attr, self.__dict__[attr_dict[attr]])
        self.df = self.df.convert_dtypes()

    def __repr__(self):
        name = self.__class__.__name__
        repr_ = f"{name}(nr={self.broid}, end={self.enddepth}, quality={self.quality})"
        return repr_

    @property
    def ns(self):
        ns = {
            None: "http://www.broservices.nl/xsd/brocommon/3.0",
            "ns6": "http://www.opengis.net/swe/2.0",
            "ns5": "http://www.broservices.nl/xsd/srcommon/1.0",
            "ns8": "http://www.broservices.nl/xsd/bhrcommon/2.0",
            "ns7": "http://www.opengis.net/om/2.0",
            "ns2": "http://www.opengis.net/gml/3.2",
            "ns4": "http://www.broservices.nl/xsd/dsbhr/2.0",
            "ns3": "http://www.w3.org/1999/xlink",
        }
        return ns

    def __set_namespaces(self):
        for k, v in self.ns.items():
            v = "{" + v + "}"
            if k is None:
                setattr(self, "defns", v)
            else:
                setattr(self, k, v)

    @property
    def attr_paths(self):
        """
        Return a dictionary of attributes in the xml and corresponding datapaths.

        """
        attr_paths = {e.tag.split("}")[1]: e.tag for e in self.data}
        return attr_paths

    @property
    def header(self):
        index = [
            "nr",
            "x",
            "y",
            "surface",
            "end",
            "quality",
            "crs",
            "reference_level",
            "codegroup",
            "soilclass",
            "textclass",
            "textprofile",
            "ca_profile",
            "reworking",
            "gw_class",
        ]

        header = pd.Series(
            [
                self.broid,
                self.x,
                self.y,
                self.z,
                self.enddepth,
                self.quality,
                self.crs,
                self.reference_level,
                self.codegroup,
                self.soilclass,
                self.textclass,
                self.textprofile,
                self.ca_profile,
                self.reworking,
                self.gw_class,
            ],
            index=index,
        )
        return header

    @property
    def attrs(self):
        return list(self.attr_paths.keys())

    def get_main_element(self, attr):
        e = self.data.find(f"{self.attr_paths[attr]}")
        return e

    def __get_data(self):
        self.data = self.root.find(f"*/{self.ns4}BHR_O")

    def _set_header_info(self):
        self.broid = self.get_main_element("broId").text
        self.quality = self.get_main_element("qualityRegime").text

        self.rd_location()
        self.get_z()
        self.bored_interval()
        self.parse_soil_classification_data()

    @staticmethod
    def get_crs(loc):
        crs = re.search(r"EPSG::(?P<crs>\d+)", str(loc.values())).group("crs")
        return crs

    @property
    def _borehole_data(self):
        data = self.get_main_element("boreholeSampleDescription")
        return data

    def rd_location(self):
        """
        Return a namedtuple of the location in RD-coordinates.

        """
        rd_element = self.get_main_element("deliveredLocation")
        loc = rd_element.find(f"{self.ns8}location")
        self.crs = self.get_crs(loc)

        x, y = loc.find(f"{self.ns2}pos").text.split(" ")
        self.x = round(float(x), 0)
        self.y = round(float(y), 0)

        return RDCoord(self.x, self.y, int(self.crs))

    def latlon_location(self):
        """
        Return a namedtuple of the location in decimal degree coordinates.

        """
        latlon_element = self.get_main_element("standardizedLocation")
        loc = latlon_element.find(f"{self.defns}location")

        crs = self.get_crs(loc)

        lat, lon = loc.find(f"{self.ns2}pos").text.split(" ")

        return DDCoord(float(lat), float(lon), int(crs))

    def get_z(self):
        elem = self.get_main_element("deliveredVerticalPosition")

        z = elem.find(f"{self.ns8}offset").text
        ref = elem.find(f"{self.ns8}verticalDatum").text

        self.z = float(z)
        self.reference_level = ref

    def bored_interval(self):
        elem = self.get_main_element("boring")

        self.begindepth = float(elem.find(f"*/{self.ns8}beginDepth").text)
        self.enddepth = self.z - float(elem.find(f"*/{self.ns8}endDepth").text)

    def parse_soillayer_element(self, layer):
        """
        Parse a soillayer element from the xml element tree of a Pedological
        soil core.

        Parameters
        ----------
        layer : lxml.etree._Element
            Soillayer element to parse.

        Returns
        -------
        l : namedtuple
            Namedtuple of the information and values of the layer.

        """
        return LayerSoilCore(
            float(layer.find(f"{self.ns8}upperBoundary").text),
            float(layer.find(f"{self.ns8}lowerBoundary").text),
            layer.find(f"*/{self.ns8}horizonCode").text,
            layer.find(f"*/*/{self.ns8}standardSoilName").text,
            layer.find(f"*/*/{self.ns8}pedologicalSoilName").text,
            # TODO:Not all of the below are in the xml
            # layer.find(f"*/*/{self.ns8}organicMatterClass").text,
            # layer.find(f"*/*/{self.ns8}carbonateClass").text,
            # layer.find(f"*/*/{self.ns8}containsGravel").text,
            # layer.find(f"*/*/{self.ns8}containsShellMatter").text,
            # float(layer.find(f"*/*/*/{self.ns8}clayContent").text),
        )

    def parse_layers(self):
        layers = self._borehole_data.find(f"{self.ns8}result")
        for l in layers.iterfind(f"{self.ns8}soilLayer"):  # noqa: E741
            yield self.parse_soillayer_element(l)

    def parse_soil_classification_data(self):
        soilclass = self._borehole_data.find(f"{self.ns8}soilClassification")

        self.codegroup = soilclass.find(f"{self.ns8}codeGroup").text
        self.soilclass = soilclass.find(f"{self.ns8}soilClass").text
        # TODO: Not all of the below parameters are always in the xml
        # self.textclass = soilclass.find(f"{self.ns8}textureClass").text
        # self.textprofile = soilclass.find(f"{self.ns8}textureProfile").text
        # self.ca_profile = soilclass.find(f"{self.ns8}carbonateProfile").text
        self.textclass = None
        self.textprofile = None
        self.ca_profile = None
        self.reworking = soilclass.find(f"{self.ns8}reworkingClass").text
        self.gw_class = soilclass.find(f"{self.ns8}groundwaterTableClass").text


class BorisXML:
    """
    Class with properties and reading/parsing methods for BORIS XML exports. BORIS is
    software developed by TNO to describe boreholes in the lab. The exported XML-format
    bears no resemblance to DINO or BRO XML standards.

    The reader is naive, which means that all data is taken from the XML as is, without
    running any checks on the parsed data.

    Note that currently only essential items to construct a BoreholeCollection are parsed.
    """

    attr_type_to_dtype = dict(
        code=str,
        median=float,
        percentage=float,
    )

    def __init__(self, xml: str | Path | etree._Element):
        if isinstance(xml, (str, Path)):
            self.root = etree.parse(xml).getroot()
        elif isinstance(xml, etree._Element):
            self.root = xml
        else:
            raise TypeError(
                "Input datatype for reader not allowed. Must be an lxml Element "
                "tree or path to an xml file."
            )

        self.layer_dataframes = list()

        for point_survey in self.root.iterchildren():
            self.parse_pointsurvey(point_survey)
        self.layer_dataframe = pd.concat(
            self.layer_dataframes, ignore_index=True
        ).convert_dtypes(convert_integer=False)

    def parse_pointsurvey(self, point_survey: etree._Element) -> None:
        # Get header and layer data as dictionaries
        header_data = self.parse_headerdata(point_survey)
        layer_dataframe = self.parse_layerdata(
            point_survey.findall("borehole/lithoDescr/lithoInterval")
        )
        layer_dataframe.insert(
            1,
            "top",
            [0] + list(layer_dataframe["bottom"].iloc[:-1].values),
        )
        layer_dataframe["end"] += header_data["surface"]

        # Repeat header data for every described borehole interval
        for key, value in header_data.items():
            layer_dataframe.insert(0, key, np.full(len(layer_dataframe), value))

        # Append data of this borehole to
        self.layer_dataframes.append(layer_dataframe)

    def parse_headerdata(self, point_survey: etree._Element) -> dict:
        # Find elements that contain header data
        identification_element = point_survey.find("identification")
        location_element = point_survey.find("surveyLocation")
        elevation_element = point_survey.find("surfaceElevation/elevation")

        # Get positional reference system attributes
        # self.horizontal_reference = location_element.find("coordinates").get(
        #     "coordSystem"
        # )
        # self.vertical_reference = elevation_element.get("levelReference") or "NAP"

        # Get header data. Note that the 'end' column is later retrieved and added to
        # the header based on the last layer interval.
        nr = safe_coerce(identification_element.get("id"), str)
        x = safe_coerce(location_element.find("coordinates/coordinateX").text, float)
        y = safe_coerce(location_element.find("coordinates/coordinateY").text, float)
        surface = safe_coerce(elevation_element.get("levelValue"), float)

        # Create header data dict
        header_data = dict(surface=surface, y=y, x=x, nr=nr)
        return header_data

    def parse_layerdata(self, borehole_element: etree._Element) -> pd.DataFrame:
        *_, last = borehole_element
        end = safe_coerce(last.get("baseDepth"), float) * -1

        num_of_layers = len(borehole_element)
        layer_data = list()
        unique_tags = ["bottom"]
        for interval in borehole_element:
            interval_dict = dict()
            interval_dict["bottom"] = float(interval.get("baseDepth"))
            for data_element in interval.getchildren():
                if data_element.tag not in unique_tags:
                    unique_tags += [data_element.tag]
                if data_element.attrib:
                    attr_type, attr_value = (
                        data_element.attrib.keys()[0],
                        data_element.attrib.values()[0],
                    )
                    interval_dict[data_element.tag] = safe_coerce(
                        attr_value, self.attr_type_to_dtype[attr_type]
                    )
                else:
                    interval_dict[data_element.tag] = safe_coerce(
                        data_element.text, str
                    )
            layer_data.append(interval_dict)

        # Construct final dataframe
        layer_df = pd.DataFrame(columns=unique_tags, index=range(num_of_layers))
        for i, layer in enumerate(layer_data):
            layer_df.update(pd.DataFrame(layer, index=[i]))
        layer_df.insert(0, "end", np.full(num_of_layers, end))
        return layer_df



from pathlib import Path
from typing import Any, Callable

import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj import CRS

from geost.base import (
    BoreholeCollection,
    Collection,
    CptCollection,
    DiscreteData,
    LayeredData,
    PointHeader,
)
from geost.bro import BroApi
from geost.io import _parse_cpt_gef_files
from geost.io.parsers import BorisXML, SoilCore
from geost.utils import dataframe_to_geodataframe

MANDATORY_LAYERED_DATA_COLUMNS = ["nr", "x", "y", "surface", "end", "top", "bottom"]
MANDATORY_DISCRETE_DATA_COLUMNS = ["nr", "x", "y", "surface", "end", "depth"]
MANDATORY_INCLINED_LAYERED_DATA_COLUMNS = [
    "nr",
    "x",
    "y",
    "x_bot",
    "y_bot",
    "surface",
    "end",
    "top",
    "bottom",
]
GEOMETRY_TO_SELECTION_FUNCTION = {
    "Polygon": "select_within_polygons",
    "Line": "select_with_lines",
    "Point": "select_with_points",
}


LLG_COLUMN_MAPPING = {
    "BOORP": "nr",
    "XCO": "x",
    "YCO": "y",
    "MV_HoogteNAP": "surface",
    "BoringEinddiepteNAP": "end",
    "BEGIN_DIEPTE": "top",
    "EIND_DIEPTE": "bottom",
}


def __read_file(file: str | Path, **kwargs) -> pd.DataFrame:
    """
    Read parquet file.

    Parameters
    ----------
    file : Path
        Path to file to be read.
    kwargs : optional
        Kwargs to pass to the read function

    Returns
    -------
    pd.DataFrame
        Dataframe with contents of 'file'.

    Raises
    ------
    TypeError
        if 'file' has no valid suffix.

    """
    file = Path(file)
    suffix = file.suffix
    if suffix in [".parquet", ".pq"]:
        return pd.read_parquet(file, **kwargs)
    elif suffix in [".csv"]:
        return pd.read_csv(file, **kwargs)
    elif suffix in [".xls", ".xlsx"]:  # pragma: no cover
        return pd.read_excel(
            file, **kwargs
        )  # No cover: Excel file io fails in windows CI pipeline
    else:
        raise TypeError(
            f"Expected parquet file (with .parquet or .pq suffix) but got {suffix} file"
        )


def adjust_z_coordinates(data_dataframe: pd.DataFrame) -> pd.DataFrame:
    """
    Interpret data containing elevation or z-coordinates. GeoST data objects require
    that layer top/bottom or discrete data point z-coordinates to be increasing
    downward, starting at at 0. This function detects how elevation is currently defined
    and turns it into the desired format if required.

    Parameters
    ----------
    data_dataframe : pd.DataFrame
        Dataframe with layer or discrete data that includes columns referecing layer
        top, bottoms or discrete data z-coordinates.
    """
    top_column_label = [
        col for col in data_dataframe.columns if col in ["top", "depth"]
    ][0]
    has_bottom_column = "bottom" in data_dataframe.columns

    # TODO: think about detection. Only considers first two indices to determine
    # downward decreasing or increasing of z-coordinates.
    first_surface = data_dataframe["surface"].iloc[0]
    first_top = data_dataframe[top_column_label].iloc[0]
    second_top = data_dataframe[top_column_label].iloc[1]

    if first_top == first_surface:
        data_dataframe[top_column_label] -= data_dataframe["surface"]
        if has_bottom_column:
            data_dataframe["bottom"] -= data_dataframe["surface"]
    if first_top > second_top:
        data_dataframe[top_column_label] *= -1
        if has_bottom_column:
            data_dataframe["bottom"] *= -1

    return data_dataframe


def _check_mandatory_column_presence(
    df: pd.DataFrame, mandatory_cols: list, column_mapper: dict = None
) -> pd.DataFrame:
    missing_mandatory = [col for col in mandatory_cols if col not in df.columns]

    if missing_mandatory:
        if not column_mapper:
            column_mapper = dict()
            print(
                f"\nMandatory columns {missing_mandatory} are missing in the "
                f"data table. Columns present are: \n{list(df.columns)}\n"
            )
            for missing in missing_mandatory:
                name = input(f"Which column is {missing}? ")
                column_mapper[name] = missing

        df.rename(columns=column_mapper, inplace=True)
        df = df.loc[:, ~df.columns.duplicated(keep="last")]

    return df


def read_borehole_table(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    has_inclined: bool = False,
    as_collection: bool = True,
    column_mapper: dict = None,
    **kwargs,
) -> BoreholeCollection:
    """
    Read tabular borehole information. This is a file (parquet, csv or Excel) that
    includes row data for each (borehole) layer and must at least include the following
    column headers:

    - **nr** : Object id
    - **x** : X-coordinates according to the given horizontal_reference
    - **y** : Y-coordinates according to the given horizontal_reference
    - **surface** : Surface elevation according to the given vertical_reference
    - **end** : End depth according to the given vertical_reference
    - **top** : Top depth of each layer
    - **bottom** : Bottom depth of each layer

    In case there are inclined boreholes in the layer data, you additionally require:

    - **x_bot** : X-coordinates of the layer bottoms
    - **y_bot** : Y-coordinates of the layer bottoms

    If you are reading a file that uses different column names, see the optional
    argument "column_mapper" below. There are no further limits or requirements to
    additional (data) columns.

    Parameters
    ----------
    file : str | Path
        Path to file to be read. Depending on the file extension, the corresponding
        Pandas read function will be called. This can be either pandas.read_parquet,
        pandas.read_csv or pandas.read_excel. The **kwargs that can be given to this
        function will be passed to the selected Panda's read function.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    has_inclined : bool, optional
        If True, the borehole data table contains inclined boreholes. The default is False.
    as_collection : bool, optional
        If True, the borehole table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.LayeredData` object is returned. The default is True.
    column_mapper: dict, optional
        If the file to be read uses different column names than the ones given in the
        description above, you can use a dictionary mapping to translate the column
        names to the required format. column_mapper is None by default, in which case
        the user is asked for input if not all required columns are encountered in the
        file.
    kwargs: optional
        Optional keyword arguments for Pandas.read_parquet, Pandas.read_csv or
        Pandas.read_excel.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        Instance of :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        depending on if the table is read as a collection or not.

    Examples
    --------
    Suppose we have a file of boreholes with columns 'nr', 'x', 'y', 'maaiveld', 'end',
    'top' and 'bottom'. The column 'maaiveld' corresponds to the required column
    'surface' in GeoST and thus needs to be remapped. The x and y-coordinates are given
    in WGS84 UTM 31N whereas the vertical datum is given in the Belgian Ostend height
    vertical datum:

    >>> read_borehole_table(file, 32631, 'Ostend height', column_mapper={'maaiveld': 'surface'})

    """
    boreholes = __read_file(file, **kwargs)

    if has_inclined:
        boreholes = _check_mandatory_column_presence(
            boreholes, MANDATORY_INCLINED_LAYERED_DATA_COLUMNS, column_mapper
        )
    else:
        boreholes = _check_mandatory_column_presence(
            boreholes, MANDATORY_LAYERED_DATA_COLUMNS, column_mapper
        )
    # Figure out top and bottom reference and coerce to downward increasing from 0 if
    # required.
    boreholes = adjust_z_coordinates(boreholes)
    boreholes = LayeredData(boreholes, has_inclined=has_inclined)

    if as_collection:
        boreholes = boreholes.to_collection(horizontal_reference, vertical_reference)

    return boreholes


def read_cpt_table(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    as_collection: bool = True,
    column_mapper: dict = None,
    **kwargs,
) -> CptCollection:
    """
    Read tabular CPT information. This is a file (parquet, csv or Excel) that includes
    row data for each CPT depth interval and must at least include the following header
    information as columns:

    nr      - Object id
    x       - X-coordinates according to the given horizontal_reference
    y       - Y-coordinates according to the given horizontal_reference
    surface - Surface elevation according to the given vertical_reference
    end     - End depth according to the given vertical_reference
    depth   - Depth in meters below the surface

    Parameters
    ----------
    file : str | Path
        File with the CPT information to read.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    as_collection : bool, optional
        If True, the CPT table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.DiscreteData` object is returned. The default is True.
    column_mapper: dict, optional
        If the file to be read uses different column names than the ones given in the
        description above, you can use a dictionary mapping to translate the column
        names to the required format. column_mapper is None by default, in which case
        the user is asked for input if not all required columns are encountered in the
        file.
    kwargs: optional
        Optional keyword arguments for Pandas.read_parquet, Pandas.read_csv or
        Pandas.read_excel.

    Returns
    -------
    :class:`~geost.base.CptCollection`
        Instance of :class:`~geost.base.CptCollection`.

    """
    cpts = __read_file(file, **kwargs)

    cpts = _check_mandatory_column_presence(
        cpts, MANDATORY_DISCRETE_DATA_COLUMNS, column_mapper
    )
    cpts = DiscreteData(cpts)

    if as_collection:
        cpts = cpts.to_collection(horizontal_reference, vertical_reference)

    return cpts


def read_nlog_cores(file: str | Path) -> BoreholeCollection:
    """
    Read NLog boreholes from the 'nlog_stratstelsel' Excel file. You can find this
    distribution of borehole data here: https://www.nlog.nl/boringen

    Warning: reading this Excel file is really slow (~10 minutes). Converting it to
    parquet using :func:`~geost.utils.excel_to_parquet` once and using that file instead
    allows for much faster reading of nlog data.

    Parameters
    ----------
    file : str | Path
        Path to nlog_stratstelsel.xlsx or .parquet

    Returns
    -------
    :class:`~geost.borehole.BoreholeCollection`
        :class:`~geost.borehole.BoreholeCollection`
    """
    nlog_cores = __read_file(file)

    nlog_cores.rename(
        columns={
            "NITG_NR": "nr",
            "X_TOP_RD": "x",
            "Y_TOP_RD": "y",
            "X_BOTTOM_RD": "x_bot",
            "Y_BOTTOM_RD": "y_bot",
            "TV_TOP_NAP": "top",
            "TV_BOTTOM_NAP": "bottom",
        },
        inplace=True,
    )

    nlog_cores["top"] *= -1
    nlog_cores["bottom"] *= -1

    nrs = []
    x = []
    y = []
    x_bot = []
    y_bot = []
    surface = []
    end = []
    nrs_data = []
    surface_data = []
    end_data = []
    for nr, data in nlog_cores.groupby("nr"):
        nrs.append(nr)
        x.append(data["x"].iloc[0])
        y.append(data["y"].iloc[0])
        x_bot.append(data["x_bot"].iloc[0])
        y_bot.append(data["y_bot"].iloc[0])
        surface.append(data["top"].iloc[0])
        end.append(data["bottom"].iloc[-1])
        nrs_data += [nr for i in range(len(data))]
        surface_data += [data["top"].iloc[0] for i in range(len(data))]
        end_data += [data["bottom"].iloc[-1] for i in range(len(data))]

    surface_end_df = pd.DataFrame(
        {"nr": nrs_data, "surface": surface_data, "end": end_data}
    ).drop_duplicates("nr")

    nlog_cores = nlog_cores.merge(surface_end_df, on="nr", how="left")
    nlog_cores = adjust_z_coordinates(nlog_cores)

    collection = LayeredData(nlog_cores, has_inclined=True).to_collection(28992, 5709)

    return collection


def read_xml_boris(
    file: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    as_collection: bool = True,
) -> BoreholeCollection:
    """
    Read export XML of the BORIS software. BORIS is software developed by TNO to
    describe boreholes in the lab. The exported XML-format bears no resemblance to DINO
    or BRO XML standards

    Parameters
    ----------
    file : str | Path
        File with the borehole information to read.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.
    as_collection : bool, optional
        If True, the CPT table will be read as a :class:`~geost.base.Collection`
        which includes a header object and spatial selection functionality. If False,
        a :class:`~geost.base.LayeredData` object is returned. The default is True.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        Instance of :class:`~geost.base.BoreholeCollection` or :class:`~geost.base.LayeredData`
        depending on if the table is read as a collection or not.
    """
    boris_data = BorisXML(file)
    boreholes = LayeredData(boris_data.layer_dataframe, has_inclined=False)

    if as_collection:
        # Think of a better way to translate non-standard BORIS crs encoding or keep it
        # user-defined (like we do in other reader functions)
        boreholes = boreholes.to_collection(horizontal_reference, vertical_reference)

    return boreholes


def read_xml_geotechnical_cores(
    file_or_folder: str | Path,
) -> BoreholeCollection:  # pragma: no cover
    """
    NOTIMPLEMENTED
    Read xml files of BRO geotechnical boreholes (IMBRO or IMBRO/A quality).
    Decribed in NEN14688 standards

    """
    pass


def read_xml_soil_cores(
    file_or_folder: str | Path,
) -> BoreholeCollection:  # pragma: no cover
    """
    NOTIMPLEMENTED
    Read xml files of BRO soil boreholes (IMBRO or IMBRO/A quality).

    """
    pass


def read_xml_geological_cores(
    file_or_folder: str | Path,
) -> BoreholeCollection:  # pragma: no cover
    """
    NOTIMPLEMENTED
    Read xml files of DINO geological boreholes.

    """
    pass


def read_gef_cores(
    file_or_folder: str | Path,
) -> BoreholeCollection:  # pragma: no cover
    """
    NOTIMPLEMENTED
    Read gef files of boreholes.

    """
    pass


def read_gef_cpts(file_or_folder: str | Path) -> CptCollection:
    """
    Read one or more GEF files of CPT data into a geost CptCollection.

    Parameters
    ----------
    file_or_folder : str | Path
        GEF files to read.

    Returns
    -------
    :class:`~geost.base.CptCollection`
        :class:`~geost.base.CptCollection` of the GEF file(s).

    """
    data = _parse_cpt_gef_files(file_or_folder)
    df = pd.concat(data)

    return DiscreteData(df).to_collection()


def read_xml_cpts(file_or_folder: str | Path) -> CptCollection:  # pragma: no cover
    """
    NOTIMPLEMENTED
    Read xml files of cpts.

    """
    pass


def get_bro_objects_from_bbox(
    object_type: str,
    xmin: int | float,
    xmax: int | float,
    ymin: int | float,
    ymax: int | float,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
) -> BoreholeCollection:
    """
    Directly download objects from the BRO to a geost collection object based on the
    given bounding box

    Parameters
    ----------
    object_type : str
        BRO object type to retrieve: BHR-GT, BHR-P, BHR-G or BHR-CPT
    xmin : int | float
        minimal x-coordinate of the bounding box
    xmax : int | float
        maximal x-coordinate of the bounding box
    ymin : int | float
        minimal y-coordinate of the bounding box
    ymax : int | float
        maximal y-coordinate of the bounding box
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    BoreholeCollection
        Instance of either :class:`~geost.borehole.BoreholeCollection` or
        :class:`~geost.borehole.CptCollection` containing only objects selected by
        this method.
    """
    api = BroApi()
    api.search_objects_in_bbox(
        xmin=xmin,
        xmax=xmax,
        ymin=ymin,
        ymax=ymax,
        object_type=object_type,
    )
    bro_objects = api.get_objects(api.object_list, object_type=object_type)
    bro_parsed_objects = []
    # TODO: The below has to be adjusted for different BRO objects
    for bro_object in bro_objects:
        try:
            object = SoilCore(bro_object)
        except (TypeError, AttributeError) as err:
            print("Cant read a soil core")
            print(err)
        bro_parsed_objects.append(object.df)

    dataframe = pd.concat(bro_parsed_objects).reset_index()

    collection = LayeredData(dataframe, has_inclined=False).to_collection(
        horizontal_reference=28992, vertical_reference=5709
    )
    collection.change_horizontal_reference(horizontal_reference)
    collection.change_vertical_reference(vertical_reference)

    return collection


def get_bro_objects_from_geometry(
    object_type: str,
    geometry_file: Path | str,
    buffer: int | float = 0,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
) -> BoreholeCollection:
    """
    Directly download objects from the BRO to a geost collection object based on given
    geometries such as points, lines or polygons.

    Parameters
    ----------
    object_type : str
        BRO object type to retrieve: BHR-GT, BHR-P, BHR-G or BHR-CPT
    geometry_file : Path | str
        Path to geometry file. i.e a point/line/polygon shapefile/geopackage/geoparquet
    buffer : int | float, optional
        Buffer distance. Required to be larger than 0 for line and point geometries.
        Adds an extra buffer zone around a polygon to select from, by default 0
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    BoreholeCollection
        Instance of either :class:`~geost.borehole.BoreholeCollection` or
        :class:`~geost.borehole.CptCollection` containing only objects selected by
        this method.
    """
    file = Path(geometry_file)
    if file.suffix == ".parquet":
        geometry = gpd.read_parquet(geometry_file)
    else:
        geometry = gpd.read_file(geometry_file)
    api = BroApi()
    api.search_objects_in_bbox(
        xmin=geometry.bounds.minx.values[0],
        xmax=geometry.bounds.maxx.values[0],
        ymin=geometry.bounds.miny.values[0],
        ymax=geometry.bounds.maxy.values[0],
        object_type=object_type,
    )
    bro_objects = api.get_objects(api.object_list, object_type=object_type)
    bro_parsed_objects = []
    # TODO: The below has to be adjusted for different BRO objects
    for i, bro_object in enumerate(bro_objects):
        try:
            object = SoilCore(bro_object)
            print(i)
        except (TypeError, AttributeError) as err:
            print("Cant read a soil core")
            print(err)
        bro_parsed_objects.append(object.df)

    dataframe = pd.concat(bro_parsed_objects).reset_index()

    collection = LayeredData(dataframe, has_inclined=False).to_collection(
        horizontal_reference=28992, vertical_reference=5709
    )
    collection = getattr(collection, GEOMETRY_TO_SELECTION_FUNCTION[geometry.type[0]])(
        geometry, buffer
    )
    collection.change_horizontal_reference(horizontal_reference)
    collection.change_vertical_reference(vertical_reference)

    return collection


def read_uullg_tables(
    header_table: str | Path,
    data_table: str | Path,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
    **kwargs,
) -> BoreholeCollection:
    """
    Read the header and data tables from UU-LLG boreholes (see: EasyDans KNAW) from a
    local csv or parquet file into a GeoST BoreholeCollection object.

    Reference dataset: https://easy.dans.knaw.nl/ui/datasets/id/easy-dataset:74935
    See https://wikiwfs.geo.uu.nl/ for additional information.

    Parameters
    ----------
    header_table : str | Path
        Path to file of the UU-LLG header table.
    data_table : str | Path
        Path to file of the UU-LLG data table.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    :class:`~geost.base.BoreholeCollection`
        Instance of :class:`~geost.base.BoreholeCollection` of the UU-LLG data.

    """
    header = __read_file(header_table, **kwargs)
    data = __read_file(data_table, **kwargs)

    header.rename(columns=LLG_COLUMN_MAPPING, inplace=True)
    data.rename(columns=LLG_COLUMN_MAPPING, inplace=True)

    header = header.astype({"nr": str})
    data = data.astype({"nr": str})

    header = dataframe_to_geodataframe(header).set_crs(horizontal_reference)

    add_header_cols = [c for c in ["x", "y", "surface", "end"] if c not in data.columns]
    if add_header_cols:
        data = data.merge(header[["nr"] + add_header_cols], on="nr", how="left")

    header = PointHeader(header, vertical_reference)
    data = LayeredData(data)
    return BoreholeCollection(header, data)


def read_collection_geopackage(
    filepath: str | Path,
    collection_type: Collection,
    horizontal_reference: str | int | CRS = 28992,
    vertical_reference: str | int | CRS = 5709,
):
    """
    Read a GeoST stored Geopackage file of `geost.base.Collection` objects. The
    Geopackage contains the Collection's "header" and "data" attributes as layers which
    are read into the specified Collection.

    Parameters
    ----------
    filepath : str | Path
        GeoST stored Geopackage file of a `geost.base.Collection` object.
    collection_type : Collection
        Type of GeoST Collection object the data needs to be. Subclasses of `Collection`
        (e.g. :class:`geost.base.BoreholeCollection`, :class:`geost.base.CptCollection`)
        for the available types.
    horizontal_reference : str | int | CRS, optional
        EPSG of the data's horizontal reference. Takes anything that can be interpreted
        by pyproj.crs.CRS.from_user_input(). The default is 28992.
    vertical_reference : str | int | CRS, optional
        EPSG of the data's vertical datum. Takes anything that can be interpreted by
        pyproj.crs.CRS.from_user_input(). However, it must be a vertical datum. FYI:
        "NAP" is EPSG 5709 and The Belgian reference system (Ostend height) is ESPG
        5710. The default is 5709.

    Returns
    -------
    Subclass of `geost.base.Collection`
        Subclass which is specified as collection_type.

    Raises
    ------
    ValueError
        Raises ValueError if the collection_type is not supported.

    Examples
    --------
    Any GeoST `Collection` object can be stored as a geopackage file:

    >>> original_collection = geost.data.boreholes_usp()
    >>> original_collection
    BoreholeCollection:
    # header = 67
    >>> original_collection.to_geopackage("./collection.gpkg")

    Reading the stored collection using `geost.read_collection_geopackage` returns the
    stored collection:

    >>> collection = geost.read_collection_geopackage(
    ...     "./collection.gpkg", collection_type=BoreholeCollection
    ...     )
    >>> collection
    BoreholeCollection:
    # header = 67

    """
    # TODO: Check collection type inference possibility from geopackage metadata.
    if collection_type == BoreholeCollection:
        header_type = PointHeader
        data_type = LayeredData
    elif collection_type == CptCollection:
        header_type = PointHeader
        data_type = DiscreteData
    else:
        raise ValueError(f"Collection type {collection_type} not supported.")

    header = gpd.read_file(filepath, layer="header")
    header.set_crs(horizontal_reference, inplace=True, allow_override=True)
    data = gpd.read_file(filepath, layer="data")
    return collection_type(header_type(header, vertical_reference), data_type(data))


def read_pickle(filepath: str | Path, **kwargs) -> Any:
    """
    Read pickled GeoST object (or any object) from file.

    Parameters
    ----------
    filepath : str | Path
        Path to pickle file to open.
    **kwargs
        pd.read_pickle kwargs. See relevant Pandas documentation.

    Returns
    -------
    Any
        Returns the Python object that was stored in the pickle.

    Examples
    --------
    Any GeoST :class:`geost.base.Collection` object can be stored as a pickle file:

    >>> original_collection = geost.data.boreholes_usp()
    >>> original_collection
    BoreholeCollection:
    # header = 67
    >>> original_collection.to_pickle("./collection.pkl")

    Reading the stored pickle using `geost.read_pickle` returns the stored collection:

    >>> collection = geost.read_pickle("./collection.pkl")
    >>> collection
    BoreholeCollection:
    # header = 67

    """
    pkl = pd.read_pickle(filepath, **kwargs)
    return pkl



import operator
from pathlib import Path
from typing import Any, Union

import geopandas as gpd
import pandas as pd
from pyogrio.errors import FieldError
from shapely.geometry import Point

COMPARISON_OPERATORS = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}

ARITHMIC_OPERATORS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
}


def csv_to_parquet(
    file: Union[str, Path], out_file: Union[str, Path] = None, **kwargs
) -> None:
    """
    Convert csv table to parquet.

    Parameters
    ----------
    file : Union[str, Path]
        Path to csv file to convert.
    out_file : Union[str, Path], optional
        Path to parquet file to be written. If not provided it will use the path of
        'file'.
    **kwargs
        pandas.read_csv kwargs. See pandas.read_csv documentation.

    Raises
    ------
    TypeError
        If 'file' is not a csv file
    """
    file = Path(file)
    if file.suffix != ".csv":
        raise TypeError(f"File must be a csv file, but a {file.suffix}-file was given")

    df = pd.read_csv(file, **kwargs)
    if out_file is None:
        df.to_parquet(file.parent / (file.stem + ".parquet"))
    else:
        df.to_parquet(out_file)


def excel_to_parquet(
    file: Union[str, Path], out_file: Union[str, Path] = None, **kwargs
) -> None:
    """
    Convert excel table to parquet.

    Parameters
    ----------
    file : Union[str, Path]
        Path to excel file to convert.
    out_file : Union[str, Path], optional
        Path to parquet file to be written. If not provided it will use the path of
        'file'.
    **kwargs
        pandas.read_excel kwargs. See pandas.read_excel documentation.

    Raises
    ------
    TypeError
        If 'file' is not an xlsx or xls file.
    """
    file = Path(file)
    if file.suffix not in [".xlsx", ".xls"]:
        raise TypeError(
            f"File must be an excel file, but a {file.suffix}-file was given"
        )

    df = pd.read_excel(file, **kwargs)
    if out_file is None:
        df.to_parquet(file.parent / (file.stem + ".parquet"))
    else:
        df.to_parquet(out_file)


def get_path_iterable(path: Path, wildcard: str = "*"):
    if path.is_file():
        return [path]
    elif path.is_dir():
        return path.glob(wildcard)
    else:
        raise TypeError("Given path is not a file or a folder")


def safe_float(number):
    try:
        return float(number)
    except ValueError:
        return None


def warn_user(func):
    def inner(*args, **kwargs):
        print("WARNING:\n--------")
        func(*args, **kwargs)
        print("--------\n>> CONTINUING MAY LEAD TO UNEXPECTED RESULTS\n")

    return inner


def inform_user(func):
    def inner(*args, **kwargs):
        print("PLEASE NOTICE:\n--------")
        func(*args, **kwargs)
        print("--------\n")

    return inner


def dataframe_to_geodataframe(
    df: pd.DataFrame, x_col_label: str = "x", y_col_label: str = "y", crs: int = None
) -> gpd.GeoDataFrame:
    """
    Take a dataframe with columns that indicate x and y coordinates and use these to
    turn the dataframe into a geopandas GeoDataFrame with a geometry column that
    contains shapely Point geometries.

    Parameters
    ----------
    df : pd.DataFrame
        Input dataframe with columns for x and y coordinates.
    x_col_label : str
        Label of the x-coordinate column, default x-coordinate column label is 'x'.
    y_col_label : str
        Label of the y-coordinate column, default y-coordinate column label is 'y'.
    crs : int
        EPSG number as integer.

    Returns
    -------
    gpd.GeoDataFrame
        GeoDataFrame with point geometries in addition to input dataframe data.

    Raises
    ------
    IndexError
        If input dataframe does not have a valid column for 'x' or 'y'.
    """
    points = [Point([x, y]) for x, y in zip(df[x_col_label], df[y_col_label])]
    gdf = gpd.GeoDataFrame(df, geometry=points, crs=crs)
    return gdf


def save_pickle(data: Any, path: str | Path, **kwargs) -> None:
    """
    Save object as a pickle file (.pkl, .pickle) using Pandas to_pickle.

    Parameters
    ----------
    data : Any
        Object to save.
    path : str | Path
        Path to pickle file.
    **kwargs
            pd.to_pickle kwargs. See relevant Pandas documentation.

    """
    pd.to_pickle(data, path, **kwargs)


def _to_geopackage(
    data: gpd.GeoDataFrame, outfile: str | Path, error_note: str, **kwargs
):
    """
    Helper to add GeoST specific information if a Pyogrio error is raised when data is
    exported to Geopackage but contians invalid column names.

    Raises
    ------
    e
        Pyogrio error with added GeoST information when data has invalid column names.
    """
    try:
        data.to_file(outfile, **kwargs)
    except FieldError as e:
        e.add_note(f"Invalid column name in {error_note}, cannot write GPKG.")
        raise e



import pyproj
import pytest
from numpy.testing import assert_almost_equal

from geost import projections


class TestProjections:
    @pytest.mark.unittest
    def test_get_horizontal_transformer(self):
        t = projections.horizontal_reference_transformer(28992, 4326)
        assert isinstance(t, pyproj.transformer.Transformer)

    @pytest.mark.unittest
    def test_get_vertical_transformer(self):
        t = projections.vertical_reference_transformer(28992, 5709, 5710)
        assert isinstance(t, pyproj.transformer.Transformer)

    @pytest.mark.unittest
    def test_xy_to_ll(self):
        lat_dec, lon_dec = projections.xy_to_ll(141000, 455000, 28992)
        assert_almost_equal(lon_dec, 5.182956740521864)
        assert_almost_equal(lat_dec, 52.083091710433095)



from typing import Iterable, List, TypeVar

import numpy as np
import pandas as pd
import pyvista as pv
from pyvista import MultiBlock


def prepare_borehole(borehole: pd.DataFrame, vertical_factor: float) -> np.ndarray:
    borehole_xyz = borehole[["x", "y", "bottom"]].to_numpy()
    surface_xyz = np.array(
        [borehole_xyz[0, 0], borehole_xyz[0, 1], borehole["surface"].iloc[0]]
    )
    borehole_xyz = np.vstack([surface_xyz, borehole_xyz], dtype="float64")
    borehole_xyz[:, 2] *= vertical_factor
    return borehole_xyz


def generate_cylinders(
    table: pd.DataFrame,
    data_columns: List[str],
    radius: float,
    vertical_factor: float,
) -> Iterable:
    boreholes = table.groupby("nr")
    for _, borehole in boreholes:
        borehole_prepared = prepare_borehole(borehole, vertical_factor)
        poly = pv.PolyData(borehole_prepared)
        line_segments = np.arange(0, len(borehole_prepared), dtype=np.int_)
        line_segments = np.insert(line_segments, 0, len(borehole_prepared))
        poly.lines = line_segments

        for data_column in data_columns:
            poly[data_column] = np.hstack(
                [borehole[data_column].values[0], borehole[data_column].values]
            )
        cylinder = poly.tube(radius=radius)
        yield cylinder


def borehole_to_multiblock(
    table: pd.DataFrame,
    data_columns: List[str],
    radius: float,
    vertical_factor: float,
) -> MultiBlock:
    """
    Create a PyVista MultiBlock object from the parsed boreholes/cpt's.

    Parameters
    ----------
    table : pd.DataFrame
        Table of borehole/CPT objects. This is CptCollection.data or
        BoreholeCollection.data.
    data_columns : List[str]
        Column names of data arrays to write in the vtk file
    radius : float
        Radius of borehole cylinders
    vertical_factor : float
        Vertical adjustment factor to convert e.g. heights in cm to m.

    Returns
    -------
    pv.MultiBlock
        MultiBlock object with boreholes represented as cylinder geometries

    """
    cylinders = generate_cylinders(table, data_columns, radius, vertical_factor)
    cylinders_multiblock = pv.MultiBlock(list(cylinders))
    return cylinders_multiblock
